// src/lib.rs

use pyo3::prelude::*;
use pyo3::types::{PyAny, PyDict};
use numpy::{PyArray1, PyArray2};
use ndarray::{Array1, Array2, ArrayView1, ArrayView2, Axis};
use statrs::distribution::{ContinuousCDF, Normal};

#[derive(Debug)]
struct MargEffError(String);

impl From<pyo3::PyErr> for MargEffError {
    fn from(e: pyo3::PyErr) -> MargEffError {
        MargEffError(format!("Python error: {e}"))
    }
}

impl From<MargEffError> for PyErr {
    fn from(e: MargEffError) -> PyErr {
        PyErr::new::<pyo3::exceptions::PyRuntimeError, _>(e.0)
    }
}

#[derive(Debug, Clone, Copy, PartialEq)]
enum ModelType {
    Logit,
    Probit,
    Unknown,
}

/// Detect if statsmodels model is Logit or Probit
fn detect_model_type(model: &PyAny) -> Result<ModelType, MargEffError> {
    let model_obj = model.getattr("model").unwrap_or(model);
    let cls: String = model_obj
        .getattr("__class__")?
        .getattr("__name__")?
        .extract()?;
    let lc = cls.to_lowercase();
    if lc == "logit" {
        Ok(ModelType::Logit)
    } else if lc == "probit" {
        Ok(ModelType::Probit)
    } else {
        Ok(ModelType::Unknown)
    }
}

/// Attempt to get a 1D f64 NumPy array (view).  
/// If not a NumPy array, calls `.to_numpy()` and tries again.
fn get_array1_f64(obj: &PyAny) -> Result<Array1<f64>, MargEffError> {
    match obj.downcast::<PyArray1<f64>>() {
        Ok(pyarr) => {
            let view = unsafe { pyarr.as_array() };
            Ok(view.to_owned())
        }
        Err(_down_err) => {
            let np_obj = obj.call_method0("to_numpy")?;
            let arr2 = np_obj.downcast::<PyArray1<f64>>().map_err(|e2| {
                MargEffError(format!("downcast to PyArray1 after .to_numpy() failed: {e2}"))
            })?;
            let view = unsafe { arr2.as_array() };
            Ok(view.to_owned())
        }
    }
}

/// Attempt to get a 2D f64 NumPy array (view).  
/// If not a NumPy array, calls `.to_numpy()` and tries again.
fn get_array2_f64(obj: &PyAny) -> Result<Array2<f64>, MargEffError> {
    match obj.downcast::<PyArray2<f64>>() {
        Ok(pyarr) => {
            let view = unsafe { pyarr.as_array() };
            Ok(view.to_owned())
        }
        Err(_down_err) => {
            let np_obj = obj.call_method0("to_numpy")?;
            let arr2 = np_obj.downcast::<PyArray2<f64>>().map_err(|e2| {
                MargEffError(format!("downcast to PyArray2 after .to_numpy() failed: {e2}"))
            })?;
            let view = unsafe { arr2.as_array() };
            Ok(view.to_owned())
        }
    }
}

fn logistic_cdf(z: f64) -> f64 {
    1.0 / (1.0 + (-z).exp())
}
fn logistic_pdf(z: f64) -> f64 {
    let e = z.exp();
    e / (1.0 + e).powi(2)
}
fn normal_pdf(z: f64) -> f64 {
    (-0.5 * z * z).exp() / (2.0 * std::f64::consts::PI).sqrt()
}

/// cdf for logit or probit
fn cdf_logit_probit(is_logit: bool, z: f64) -> f64 {
    if is_logit {
        logistic_cdf(z)
    } else {
        let dist = Normal::new(0.0, 1.0).unwrap();
        dist.cdf(z)
    }
}
/// pdf for logit or probit
fn pdf_logit_probit(is_logit: bool, z: f64) -> f64 {
    if is_logit {
        logistic_pdf(z)
    } else {
        normal_pdf(z)
    }
}

/// Check if a column is strictly 0/1 => discrete
fn is_discrete_column(col: &ArrayView1<f64>) -> bool {
    col.iter().all(|&v| (v - 0.0).abs() < 1e-12 || (v - 1.0).abs() < 1e-12)
}

/// Significance stars
fn add_significance_stars(p: f64) -> &'static str {
    if p < 0.001 { "***" }
    else if p < 0.01 { "**" }
    else if p < 0.05 { "*" }
    else if p < 0.1 { "." }
    else { "" }
}

/// Extract from statsmodels => (model type, params, cov, exog_names, X)
fn extract_binary_model(model: &PyAny)
    -> Result<(ModelType, Array1<f64>, Array2<f64>, Vec<String>, Array2<f64>), MargEffError>
{
    let mtype = detect_model_type(model)?;
    if mtype == ModelType::Unknown {
        return Err(MargEffError("Expected Logit or Probit model".into()));
    }
    let model_obj = model.getattr("model")?;
    let exog_names: Vec<String> = model_obj.getattr("exog_names")?.extract()?;

    let X = get_array2_f64(model_obj.getattr("exog")?)?;
    let params = get_array1_f64(model.getattr("params")?)?;
    let cov = get_array2_f64(model.call_method0("cov_params")?)?;
    Ok((mtype, params, cov, exog_names, X))
}

/// Remove intercept from final arrays if present
fn remove_intercept(
    params: Array1<f64>,
    cov: Array2<f64>,
    mut names: Vec<String>,
    X: Array2<f64>,
) -> (Array1<f64>, Array2<f64>, Vec<String>, Array2<f64>) {
    if let Some(idx) = names.iter().position(|nm| {
        nm.eq_ignore_ascii_case("const") || nm.eq_ignore_ascii_case("intercept")
    }) {
        // physically remove
        names.remove(idx);
        let (nobs, nvars) = X.dim();
        if nvars > 1 {
            let indices: Vec<usize> = (0..nvars).filter(|&c| c != idx).collect();
            let newX = X.select(ndarray::Axis(1), &indices);
            let new_params = params.select(ndarray::Axis(0), &indices);
            let cov_indices: Vec<usize> = (0..cov.nrows()).filter(|&r| r != idx).collect();
            let new_cov = cov
                .select(ndarray::Axis(0), &cov_indices)
                .select(ndarray::Axis(1), &cov_indices);
            (new_params, new_cov, names, newX)
        } else {
            (
                Array1::zeros(0),
                Array2::zeros((0, 0)),
                names,
                Array2::zeros((nobs, 0)),
            )
        }
    } else {
        (params, cov, names, X)
    }
}

//
// partial_effects_f64 => single pass for sums
//
fn partial_effects_f64(
    is_logit: bool,
    X: &Array2<f64>,
    params: &Array1<f64>,
    disc: &[bool],
) -> Array1<f64> {
    let (nobs, k) = X.dim();
    let mut sums = vec![0.0; k];

    for i in 0..nobs {
        let row_i = X.row(i);
        let z_i = row_i.dot(params);
        for j in 0..k {
            if disc[j] {
                // difference approach
                let x_ij = row_i[j];
                let plus_val = z_i + (1.0 - x_ij)*params[j];
                let minus_val= z_i - x_ij*params[j];
                sums[j]+= cdf_logit_probit(is_logit, plus_val)
                        - cdf_logit_probit(is_logit, minus_val);
            } else {
                let pdf = pdf_logit_probit(is_logit, z_i);
                sums[j]+= params[j]* pdf;
            }
        }
    }

    // average
    let mut out= Array1::from(sums);
    out.mapv_inplace(|v| v / (nobs as f64));
    out
}

//
// partial_effects_f32 => single pass for sums
//
fn partial_effects_f32(
    is_logit: bool,
    X: &ndarray::Array2<f32>,
    params: &ndarray::Array1<f32>,
    disc: &[bool],
) -> ndarray::Array1<f32> {
    let (nobs,k)= X.dim();
    let mut sums= vec![0.0_f32; k];
    for i in 0..nobs {
        let row_i= X.row(i);
        let z_i= row_i.dot(params);
        for j in 0..k {
            if disc[j] {
                let x_ij= row_i[j];
                let plus_val= z_i + (1.0 - x_ij)* params[j];
                let minus_val= z_i - x_ij* params[j];
                sums[j]+= if is_logit {
                    logistic_cdf_f32(plus_val) - logistic_cdf_f32(minus_val)
                } else {
                    probit_cdf_f32(plus_val) - probit_cdf_f32(minus_val)
                };
            } else {
                let pdf= if is_logit {
                    logistic_pdf_f32(z_i)
                } else {
                    normal_pdf_f32(z_i)
                };
                sums[j]+= params[j]* pdf;
            }
        }
    }
    let mut out= ndarray::Array1::from(sums);
    out.mapv_inplace(|v| v/ (nobs as f32));
    out
}

// logistic cdf in f32
fn logistic_cdf_f32(z: f32) -> f32 {
    1.0_f32/(1.0_f32 + (-z).exp())
}
fn logistic_pdf_f32(z: f32) -> f32 {
    let e= z.exp();
    e/(1.0+ e).powi(2)
}
fn normal_pdf_f32(z: f32)-> f32 {
    (-0.5* z*z).exp()/(2.0* std::f32::consts::PI).sqrt()
}
fn probit_cdf_f32(z: f32)-> f32 {
    0.5_f32*(1.0_f32 + erf_approx_f32(z/(2.0_f32.sqrt())))
}
fn erf_approx_f32(x: f32)-> f32 {
    let sign= if x<0.0 { -1.0} else {1.0};
    let x_= x.abs();
    let t= 1.0_f32/(1.0_f32+ 0.3275911_f32* x_);
    let a1=0.254829592_f32;
    let a2=-0.284496736_f32;
    let a3=1.421413741_f32;
    let a4=-1.453152027_f32;
    let a5=1.061405429_f32;
    let y= 1.0_f32 - t*(a1 + t*(a2 + t*(a3 + t*(a4 + t*a5)))) * (-x_*x_).exp();
    sign* y
}

/// numeric derivative => shape(k, p_len) for f64
fn numeric_derivative_f64(
    is_logit: bool,
    X: &Array2<f64>,
    disc: &[bool],
    params: &Array1<f64>,
    cov: &Array2<f64>,
) -> (Array1<f64>, Array1<f64>) {
    let base_pe= partial_effects_f64(is_logit, X, params, disc);
    let p_len= params.len();
    let k= base_pe.len();
    if p_len==0 {
        return (base_pe, Array1::zeros(k));
    }
    let step= if !is_logit {1e-8} else {1e-6};
    let mut jac= ndarray::Array2::<f64>::zeros((k, p_len));
    let mut pcopy= params.clone();

    for pi in 0..p_len {
        pcopy[pi]+= step;
        let up_val= partial_effects_f64(is_logit, X, &pcopy, disc);
        pcopy[pi]-= 2.0* step;
        let dn_val= partial_effects_f64(is_logit, X, &pcopy, disc);
        pcopy[pi]+= step;
        jac.column_mut(pi).assign(&((&up_val - &dn_val)/(2.0* step)));
    }
    let j_c= jac.dot(cov);
    let cov_ame= j_c.dot(&jac.t());
    let se= cov_ame.diag().mapv(|v| if v>0.0 { v.sqrt()} else {0.0});
    (base_pe, se)
}

/// numeric derivative => shape(k, p_len) for f32
fn numeric_derivative_f32(
    is_logit: bool,
    X: &ndarray::Array2<f32>,
    disc: &[bool],
    params: &ndarray::Array1<f32>,
    cov: &Array2<f64>,
) -> (ndarray::Array1<f32>, Array1<f64>) {
    let base_pe= partial_effects_f32(is_logit, X, params, disc);
    let p_len= params.len();
    let k= base_pe.len();
    if p_len==0 {
        return (base_pe, Array1::zeros(k));
    }
    let step= if !is_logit {1e-8_f32} else {1e-6_f32};
    let mut jac= ndarray::Array2::<f64>::zeros((k,p_len));
    let mut pcopy= params.clone();

    for pi in 0..p_len {
        pcopy[pi]+= step;
        let up_val= partial_effects_f32(is_logit, X, &pcopy, disc);
        pcopy[pi]-= 2.0* step;
        let dn_val= partial_effects_f32(is_logit, X, &pcopy, disc);
        pcopy[pi]+= step;
        for rj in 0..k {
            let diff= (up_val[rj] - dn_val[rj])/(2.0* step);
            jac[[rj, pi]]= diff as f64;
        }
    }
    let j_c= jac.dot(cov);
    let cov_ame= j_c.dot(&jac.t());
    let se= cov_ame.diag().mapv(|v| if v>0.0 {v.sqrt()} else {0.0});
    (base_pe, se)
}

/// Build final DataFrame
fn build_dataframe(
    py: Python,
    ame_vals: &Array1<f64>,
    se_vals: &Array1<f64>,
    names: &[String],
    intercept_idx: Option<usize>,
) -> PyResult<PyObject> {
    let normal= Normal::new(0.0,1.0).unwrap();

    let mut dy_col= Vec::new();
    let mut se_col= Vec::new();
    let mut z_col= Vec::new();
    let mut p_col= Vec::new();
    let mut star_col= Vec::new();
    let mut row_names= Vec::new();

    for (j,nm) in names.iter().enumerate() {
        if Some(j)== intercept_idx {
            continue;
        }
        let dy= ame_vals[j];
        let s= se_vals[j];
        if s.abs()<1e-15 {
            dy_col.push(dy);
            se_col.push(s);
            z_col.push(0.0);
            p_col.push(1.0);
            star_col.push("");
        } else {
            let z= dy/s;
            let p= 2.0*(1.0 - normal.cdf(z.abs()));
            dy_col.push(dy);
            se_col.push(s);
            z_col.push(z);
            p_col.push(p);
            star_col.push(add_significance_stars(p));
        }
        row_names.push(nm.clone());
    }

    let data= PyDict::new(py);
    data.set_item("dy/dx", &dy_col)?;
    data.set_item("Std. Err", &se_col)?;
    data.set_item("z", &z_col)?;
    data.set_item("Pr(>|z|)", &p_col)?;
    data.set_item("Significance", &star_col)?;

    let kwargs= PyDict::new(py);
    kwargs.set_item("data", data)?;
    kwargs.set_item("index", &row_names)?;

    let pd= py.import("pandas")?;
    let df= pd.call_method("DataFrame", (), Some(kwargs))?;
    Ok(df.to_object(py))
}

#[pyfunction]
fn ame(
    py: Python,
    model: &PyAny,
    precise: Option<bool>,
) -> PyResult<PyObject> {
    // 1) Extract
    let (mtype, params, cov, names, X) = extract_binary_model(model)?;
    if mtype == ModelType::Unknown {
        return Err(MargEffError("Unsupported model".into()).into());
    }

    // 2) Remove intercept physically
    let (params, cov, names, X) = remove_intercept(params, cov, names, X);
    let is_logit= (mtype== ModelType::Logit);

    // 3) Mark discrete
    let mut disc_flags= Vec::with_capacity(X.ncols());
    for c in 0..X.ncols() {
        disc_flags.push(is_discrete_column(&X.column(c)));
    }

    // 4) Decide single/double precision numeric derivative
    let use_precise= precise.unwrap_or(true);
    if use_precise {
        let (pe,se)= numeric_derivative_f64(is_logit, &X, &disc_flags, &params, &cov);
        let intercept_idx= None;  // intercept physically removed
        build_dataframe(py, &pe, &se, &names, intercept_idx)
    } else {
        // cast to f32
        let X_f32= X.mapv(|v| v as f32);
        let p_f32= params.mapv(|v| v as f32);
        let (pe32,se_f64)= numeric_derivative_f32(is_logit, &X_f32, &disc_flags, &p_f32, &cov);
        let pe_f64= pe32.mapv(|v| v as f64);
        let intercept_idx= None; 
        build_dataframe(py, &pe_f64, &se_f64, &names, intercept_idx)
    }
}

#[pymodule]
fn febolt(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(ame, m)?)?;
    Ok(())
}

